#!/usr/bin/perl -w
use strict;
use DBI;
use threads;
use IO::Socket;
use Getopt::Std;
use POSIX;
use Time::HiRes qw( gettimeofday tv_interval usleep );
use DBI;

#
#
#  SET YOUR SQL CREDENTIALS 
#
#
my $MYDATABASE = "mpm";
my $MYSOCK = "/var/run/mysqld/mysqld.sock";
my $MYUSERNAME = "mpm";
my $MYPASSWORD = "mpm";

my %opts = ();
getopts('a:b:p:m:hdg:c:w:s:e:f:',\%opts);

my $bind_port		= "25";
my $bind_address	= "";
my $max_parallel	= 100;
my $group_connection	= 'ip';
my $greylist		= 0;
my $greylist_window	= 1800;
my $max_per_session	= 100;
my $is_debug 		= 0;
my $error_file		= "";
my @error_replies	= ();
my $error_rate		= 0;
my $drop_after		= '';

if ($opts{'a'}) {
	$drop_after = $opts{'a'};
}
if ($opts{'p'}) {
	$bind_port	= $opts{'p'};
}
if ($opts{'b'}) {
	$bind_address	= $opts{'b'};
}
if ($opts{'m'}) {
	$max_parallel	= $opts{'m'};
}
if ($opts{'g'}) {
	$greylist	= $opts{'g'};
}
if ($opts{'c'}) {
	$group_connection= $opts{'c'};
}
if ($opts{'w'}) {
	$greylist_window = $opts{'w'};
}
if ($opts{'s'}) {
	$max_per_session = $opts{'s'};
}
if ($opts{'e'}) {
	$error_rate = $opts{'e'};
}
if ($opts{'f'}) {
	$error_file = $opts{'f'};
	if (!open(ERESP, "<".$error_file)) {
		print "Couldn't open file with error responses!\n";
		exit(1);
	} else {
		while(<ERESP>) {
			$_ =~ s/^\s+//;
			$_ =~ s/\s+$//;
			push(@error_replies,$_);
			print "Error response: ".$_."\n";
		}
		close(ERESP);
	}
}
if ($opts{'d'}) {
	$is_debug = 1;
}


if ($opts{'h'}) {
	print "Usage: mpmdaemon.pl <options>
	-a <SMTP command>			close the connection after <command>
	-b <bind to IP>			IP address to bind the daemon
	-p <bind to port>		Port number to bind the daemon (25 = default)
	-m <max parallel conn.>		Allow maximum of parallel connections (100 = default)
	-g <greylist timeout>		Use greylisting - accept message after timeout seconds (0 = default/disabled)
	-w <greylist window time>	Greylist window in seconds - accept messages within x seconds after the timeout (1800 = default)
	-c <ip or group>		Count number of connection per ip or per group (ip = default)
	-s <max mails per session>	Maximum of mails within one SMTP session (100 = default)
	-e <error rate in percent>	Refuse x percent of emails with a random error
	-f <error file>			File that contains the error replies
	-d				Don't daemonize and debug
	-h				Show this help
";
	exit(0);
}

#print "Port: ".$bind_port."\n";
#print "Address: ".$bind_address."\n";


my $server = IO::Socket::INET->new(	LocalAddr => $bind_address.':'.$bind_port,
					Type => SOCK_STREAM,
					ReuseAddr => 1,
					Listen => 20000 ) or die "Couldn't be a tcp server on port $bind_port : $@\n";

if ($opts{'d'}) {
	print "$$: Bound to $bind_address Port $bind_port\n";
}

if (!$opts{'d'}) {
	chdir '/root'				or die "Can't chdir to /root: $!";
	umask 0;
	open STDIN, '/dev/null'			or die "Can't read /dev/null: $!";
	open STDOUT, '>/dev/null' 		or die "Can't write to /dev/null: $!";
	open STDERR, '>/dev/null'		or die "Can't write to /dev/null: $!";
	defined(my $pid = fork)			or die "Can't fork: $!";
	exit if $pid;
}

my $iClientPort			= "";
my $sClientAddr			= "";
my $iConnectionCount		= 0;


while (my ($client,$client_address) = $server->accept()) {
	# $client is the new connection
	my $threadInstance = threads->create("handleConnection",$client,$bind_address,$client_address,$max_parallel,$greylist,$group_connection,$greylist_window,$max_per_session,$is_debug,\@error_replies,$error_rate,$drop_after);
	$threadInstance->detach();
	if ($opts{'d'}) {
		print "$$: Connection ".$iConnectionCount." detached\n";
	}
	$iConnectionCount++;
	undef $threadInstance;
}
close($server);


sub handleConnection {
	my ($client,$bind_address,$client_address,$max_parallel,$greylist,$group_connection,$greylist_window,$max_per_session,$is_debug,$error_file,$error_rate,$drop_after) = @_;
	our $MYDATABASE;
	our $MYUSERNAME;
	our $MYPASSWORD;
	our $MYSOCK;
	threads->yield();
	fcntl($client, F_SETFL(), O_NONBLOCK());
	my $timeConnectionStart 	= [gettimeofday];
	my $timeWaitForCommand 		= 0;
	my $timeSession 		= 0;
	my $timeSessionStart 		= 0;
	my $request 			= "";
	my $isInDataMode 		= 0;
	my $iMailsSent 			= 0;
	my $iRecipients			= 0;
	my $sMailto			= "";
	my $sHelo			= "";
	my $sessionid			= time().'-'.threads->self->tid;
	my $sMailfrom			= "";
	my $greyliststatus		= 0;
	my $was_nl			= 0;

	my $dbh = DBI->connect("DBI:mysql:database=".$MYDATABASE.";mysql_socket=".$MYSOCK, $MYUSERNAME, $MYPASSWORD);
	if (!$dbh) {
		print $client "450 Connection to database failed.\n";
		exit(1);
	}


	my ($port, $ip_addr) = sockaddr_in($client_address);
	# print $sessionid.": Bound to: ".$bind_address."\n";
	#
	# pruefe auf offene Sessions
	#

	my $query = "";
	if ($group_connection eq "ip") {
		$query = "SELECT COUNT(*) AS cnt FROM mpm_sessions WHERE isopen='y' AND from_ip=".$dbh->quote(inet_ntoa($ip_addr))." AND to_ip=".$dbh->quote($bind_address);
	} else {
		$query = "SELECT COUNT(*) AS cnt FROM mpm_sessions WHERE isopen='y' AND from_ip=".$dbh->quote(inet_ntoa($ip_addr));
	}
	my $sth = $dbh->prepare($query);
	$sth->execute();
	my $ref = $sth->fetchrow_hashref();
	if ($ref->{'cnt'} >= $max_parallel) {
		$dbh->do("INSERT INTO mpm_sessions SET
				sessionid=".$dbh->quote($sessionid).",
				from_ip=".$dbh->quote(inet_ntoa($ip_addr)).",
				to_ip=".$dbh->quote($bind_address).",
				num_emails='0',
				num_recipients='0',
				duration='0',
				started=NOW(),
				ended=NOW(),
				terminationstate='MAX_SESSIONS(".$max_parallel.")',
				isopen='n'
		");
		print $client "450 Maximum parallel connections reached. Please try again later.\n";
		close($client);
		undef $client;
	} else {
		$dbh->do("INSERT INTO mpm_sessions SET
				sessionid=".$dbh->quote($sessionid).",
				from_ip=".$dbh->quote(inet_ntoa($ip_addr)).",
				to_ip=".$dbh->quote($bind_address).",
				num_emails='0',
				num_recipients='0',
				duration='0',
				started=NOW(),
				terminationstate='OPEN',
				isopen='y'
		");
		print $client "220 Mail Performance Meter ready (tid:".$sessionid." ".inet_ntoa($ip_addr).":".$port." -> ".$bind_address." Connections: ".($ref->{'cnt'}+1).") \r\n";
	}
	$sth->finish();

	while ($client) {

		$request = "";

		$timeWaitForCommand = [gettimeofday];
		while ($client && !($request = <$client>)) {
			if (	(tv_interval ($timeWaitForCommand, [gettimeofday]) > 30 && $isInDataMode == 0) ||
					(tv_interval ($timeWaitForCommand, [gettimeofday]) > 60 && $isInDataMode == 1)
			) {

				# schreibe sessionstatistik in SQL DB
				$dbh->do("UPDATE mpm_sessions SET
						num_emails='".$iMailsSent."',
						num_recipients='".$iRecipients."',
						duration=".$dbh->quote(tv_interval ($timeConnectionStart, [gettimeofday])).",
						ended=NOW(),
						terminationstate='TIMEOUT',
						isopen='n'
					WHERE sessionid=".$dbh->quote($sessionid)
				);

				print $client "450 Connection closed due to timeout.\n";
				close($client);
				undef $client;
				last;
			}
			usleep(100);
		}

		if ($request) {
			# print $request;
			if ($is_debug == 1 && $isInDataMode == 0) {
				print threads->self->tid.': <'.$request;
			}
			if ($isInDataMode == 0 && $request =~ m/^EHLO(.*?)$/i) {
				if ($drop_after eq "EHLO") {
					return;
				}
				print $client "250-mpm\r\n";
				print $client "250-SIZE 10240000\r\n";
				print $client "250-8BITMIME\r\n";
				print $client "250 DSN\r\n";
				if ($is_debug == 1) {
					print threads->self->tid.': >250-mpm'."\r\n";
					print threads->self->tid.': >250-SIZE 10240000'."\r\n";
					print threads->self->tid.': >250-8BITMIME'."\r\n";
					print threads->self->tid.': >250 DSN'."\r\n";
				}
				$sHelo = $1;
				chomp($sHelo);
				$timeSessionStart = [gettimeofday];
			} elsif ($isInDataMode == 0 && $request =~ m/^HELO(.*?)$/i) {
				if ($drop_after eq "HELO") {
					return;
				}
				print $client "250 mpm\r\n";
				if ($is_debug == 1) {
					print threads->self->tid.': >250 mpm'."\r\n";
				}
				$sHelo = $1;
				$timeSessionStart = [gettimeofday];
				chomp($sHelo);
			} elsif ($isInDataMode == 0 && $request =~ m/^MAIL FROM:(.*?)$/i) {
				if ($drop_after eq "MAIL") {
					return;
				}
				if ($sHelo eq "") {
					print $client "503 Issue EHLO/HELO first.\r\n";
					if ($is_debug == 1) {
						print threads->self->tid.': >503 Issue EHLO/HELO first.'."\r\n";
					}
				} else {
					print $client "250 Ok (Mail ".$iMailsSent.")\r\n";
					if ($is_debug == 1) {
						print threads->self->tid.': >250 Ok (Mail '.$iMailsSent.')'."\r\n";
					}
					$sMailfrom = $1;
					chomp($sMailfrom);
				}
			} elsif ($isInDataMode == 0 && $request =~ m/^NOOP/i) {
				if ($drop_after eq "NOOP") {
					return;
				}
				print $client "250 Ok\r\n";
				if ($is_debug == 1) {
					print threads->self->tid.': >250 OK'."\r\n";
				}
			} elsif ($isInDataMode == 0 && $request =~ m/^VRFY:/i) {
				if ($drop_after eq "VRFY") {
					return;
				}
				print $client "250 Ok\r\n";
				if ($is_debug == 1) {
					print threads->self->tid.': >250 OK'."\r\n";
				}
			} elsif ($isInDataMode == 0 && $request =~ m/^RCPT TO:(.*?)$/i) {
				if ($drop_after eq "RCPT") {
					return;
				}
				if ($sMailfrom eq "") {
					print $client "503 Issue MAIL FROM first.\r\n";
					if ($is_debug == 1) {
						print threads->self->tid.': >503 Issue MAIL FROM first.'."\r\n";
					}
				} else {
					$sMailto = $1;
					chomp($sMailto);
					# Pruefe Greylisting status
					if ($greylist > 0) {
						$greyliststatus = 1;
						$sth = $dbh->prepare("SELECT COUNT(*) AS cnt FROM mpm_mails WHERE
								     mailto=".$dbh->quote($sMailto)." AND
								     mailfrom=".$dbh->quote($sMailfrom)." AND
								     fromip=".$dbh->quote(inet_ntoa($ip_addr))." AND
								     greylisted='y' AND
								     created < DATE_SUB(NOW(), INTERVAL ".$greylist." SECOND) AND
								     created > DATE_SUB(NOW(), INTERVAL ".$greylist_window." SECOND)
							");
						$sth->execute();
						my $ref = $sth->fetchrow_hashref();
						if ($ref->{'cnt'} > 0) {
							$greyliststatus = 0;
						}
						$sth->finish();
					}
					if ($greyliststatus == 0) {
						if ($iRecipients >= $max_per_session) {
							print $client "450 Maximum number of recipients in this session reached.\r\n";
							if ($is_debug == 1) {
								print threads->self->tid.': >450 Maximum number of recipients in this session reached.'."\r\n";
							}
						} else {

							# Check if it was requested to refuse random mails
							my $randval = int(rand(100))+1;
							if ($error_rate > 0 && $randval < $error_rate) {
								my $randval2 = int(rand($#error_replies+1));
								print $client $error_replies[$randval2]."\r\n";
								if ($is_debug == 1) {
									print threads->self->tid.': >'.$error_replies[$randval2]."\r\n";
								}
								$dbh->do("INSERT INTO mpm_mails SET
										sessionid=".$dbh->quote($sessionid).",
										mailto=".$dbh->quote($sMailto).",
										mailfrom=".$dbh->quote($sMailfrom).",
										fromip=".$dbh->quote(inet_ntoa($ip_addr)).",
										created=NOW(),
										greylisted='n',
										refusemsg=".$dbh->quote($error_replies[$randval2])."
								");
								undef $randval2;
							} else {
								print $client "250 Ok (Mail ".$iMailsSent.")\r\n";
								if ($is_debug == 1) {
									print threads->self->tid.': >250 Ok (Mail '.$iMailsSent.')'."\r\n";
								}
								$iRecipients++;
								$dbh->do("INSERT INTO mpm_mails SET
										sessionid=".$dbh->quote($sessionid).",
										mailto=".$dbh->quote($sMailto).",
										mailfrom=".$dbh->quote($sMailfrom).",
										fromip=".$dbh->quote(inet_ntoa($ip_addr)).",
										created=NOW(),
										greylisted='n'
								");
							}
							undef $randval;
						}
					} else {
						print $client "450 Greylisted for ".$greylist." seconds\r\n";
						if ($is_debug == 1) {
							print threads->self->tid.': >450 Greylisted for '.$greylist.' seconds'."\r\n";
						}
						$dbh->do("INSERT INTO mpm_mails SET
								sessionid=".$dbh->quote($sessionid).",
								mailto=".$dbh->quote($sMailto).",
								mailfrom=".$dbh->quote($sMailfrom).",
								fromip=".$dbh->quote(inet_ntoa($ip_addr)).",
								created=NOW(),
								greylisted='y'
						");
						$sMailto = "";
					}
				}
			} elsif ($isInDataMode == 0 && $request =~ m/^DATA/i) {
				if ($drop_after eq "DATA") {
					return;
				}
				if ($sHelo eq "" || $sMailto eq "" || $sMailfrom eq "") {
					print $client "503 Issue MAIL RCPT and HELO/EHLO first.\r\n";
					if ($is_debug == 1) {
						print threads->self->tid.': >503 Issue MAIL RCPT and HELO/EHLO first.'."\r\n";
					}
				} else {
					print $client "354 End data with <CR><LF>.<CR><LF>\r\n";
					$isInDataMode = 1;
					if ($is_debug == 1) {
						print threads->self->tid.': >354 End data with <CR><LF>.<CR><LF>.'."\r\n";
					}
				}
			} elsif ($isInDataMode == 0 && $request =~ m/^RSET/i) {
				if ($drop_after eq "RSET") {
					return;
				}
				print $client "250 Ok\r\n";
				if ($is_debug == 1) {
					print threads->self->tid.': >250 OK'."\r\n";
				}
				$timeSessionStart = [gettimeofday];
				$sMailto = '';
				$sMailfrom = '';
				$greyliststatus = 0;
			} elsif ($isInDataMode == 1 && $request =~ m/^\r*\n/i) {
				$was_nl = 1;
			} elsif ($isInDataMode == 1 && $was_nl == 1 && $request =~ m/^\.\r*\n/i) {
				$timeSession = tv_interval ($timeSessionStart, [gettimeofday]);
				$isInDataMode = 0;
				$was_nl = 0;
				$iMailsSent++;
				print $client "250 Ok: queued (Mail ".$iMailsSent." ".$timeSession."s)\r\n";
				if ($is_debug == 1) {
					print threads->self->tid.': >250 Ok: queued (Mail '.$iMailsSent.' '.$timeSession.'s)'."\r\n";
				}
			} elsif ($isInDataMode == 1 && $was_nl == 1) {
				$was_nl = 0;
			} elsif ($isInDataMode == 0 && $request =~ m/^QUIT/i) {
				print $client "221 Bye (".$iMailsSent." mails sent in this session)\r\n";
				if ($is_debug == 1) {
					print threads->self->tid.': >221 Bye ('.$iMailsSent.' mails sent in this session)'."\r\n";
				}
				# schreibe sessionstatistik in SQL DB
				$dbh->do("UPDATE mpm_sessions SET
						num_emails='".$iMailsSent."',
						num_recipients='".$iRecipients."',
						duration=".$dbh->quote(tv_interval ($timeConnectionStart, [gettimeofday])).",
						ended=NOW(),
						terminationstate='QUIT',
						isopen='n'
					WHERE sessionid=".$dbh->quote($sessionid)
				);
				close($client);
				undef $client;
			}
		}
	}
#	print "Time of this connection: ".tv_interval ($timeConnectionStart, [gettimeofday])."\n";
	if ($client) {
		close($client);
		undef $client;
	}

	$dbh->disconnect();

	undef $sHelo;
	undef $sMailfrom;
	undef $request;
	undef $timeConnectionStart;
	undef $isInDataMode;
	undef $timeSession;
	undef $timeSessionStart;
	undef $timeWaitForCommand;
	undef $iMailsSent;
	undef $sessionid;
	undef $client_address;
	undef $bind_address;
	undef $port;
	undef $ip_addr;
	undef $sMailto;
	undef $dbh;
	undef $max_parallel;
	undef $greylist;
	undef $group_connection;
	undef $greylist_window;
	undef $max_per_session;
	undef $sth;
	undef $ref;
	undef $query;
	undef $greyliststatus;
	undef $is_debug;
	undef $was_nl;
}